# Orca
#
# Copyright 2014-2026 Igalia, S.L.
#
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Preferences grid for spell check support."""

from __future__ import annotations

from typing import TYPE_CHECKING

from . import guilabels, preferences_grid_base, spellcheck_presenter

if TYPE_CHECKING:
    from .spellcheck_presenter import SpellCheckPresenter


class SpellCheckPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """GtkGrid containing the Spell Check preferences page."""

    _gsettings_schema = "spellcheck"
    _documentation_summary = (
        "Use these settings to control what Orca announces while reviewing spelling and "
        "grammar suggestions."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for spell-check preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.SPELL_CHECK,
            panel_id="spellcheck_presenter.spell-check",
            description=(
                "Spell Check settings control how much detail Orca presents when "
                "interacting with supported application spell checkers.\n\n"
                + guilabels.SPELL_CHECK_DESCRIPTION
            ),
            schema="spellcheck",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SPELL_CHECK_SPELL_ERROR,
                    kind="switch",
                    summary=(
                        "Controls whether Orca spells out the current error after speaking it."
                    ),
                    schema="spellcheck",
                    key=spellcheck_presenter.SpellCheckPresenter.KEY_SPELL_ERROR,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SPELL_CHECK_SPELL_SUGGESTION,
                    kind="switch",
                    summary=(
                        "Controls whether Orca spells out the current suggestion after speaking it."
                    ),
                    schema="spellcheck",
                    key=spellcheck_presenter.SpellCheckPresenter.KEY_SPELL_SUGGESTION,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.SPELL_CHECK_PRESENT_CONTEXT,
                    kind="switch",
                    summary=(
                        "Controls whether Orca presents the surrounding text in which the "
                        "mistake occurred."
                    ),
                    schema="spellcheck",
                    key=spellcheck_presenter.SpellCheckPresenter.KEY_PRESENT_CONTEXT,
                ),
            ),
        )

    def __init__(self, presenter: SpellCheckPresenter) -> None:
        controls = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.SPELL_CHECK_SPELL_ERROR,
                getter=presenter.get_spell_error,
                setter=presenter.set_spell_error,
                prefs_key=spellcheck_presenter.SpellCheckPresenter.KEY_SPELL_ERROR,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.SPELL_CHECK_SPELL_SUGGESTION,
                getter=presenter.get_spell_suggestion,
                setter=presenter.set_spell_suggestion,
                prefs_key=spellcheck_presenter.SpellCheckPresenter.KEY_SPELL_SUGGESTION,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.SPELL_CHECK_PRESENT_CONTEXT,
                getter=presenter.get_present_context,
                setter=presenter.set_present_context,
                prefs_key=spellcheck_presenter.SpellCheckPresenter.KEY_PRESENT_CONTEXT,
            ),
        ]

        super().__init__(
            guilabels.SPELL_CHECK,
            controls,
            info_message=guilabels.SPELL_CHECK_DESCRIPTION,
        )



# Orca
#
# Copyright 2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-lines
# pylint: disable=too-many-public-methods
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-many-arguments,
# pylint: disable=too-many-positional-arguments
# pylint: disable=too-many-return-statements

"""Preferences grids for document presentation and navigation settings."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from . import (
    caret_navigator,
    document_presenter,
    guilabels,
    math_presenter,
    preferences_grid_base,
    structural_navigator,
    table_navigator,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    import gi

    from .document_presenter import DocumentPresenter

    gi.require_version("Gtk", "3.0")
    from gi.repository import Gtk


class CaretNavigationPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """Sub-grid for caret navigation settings within the Documents page."""

    _gsettings_schema = "caret-navigation"
    _documentation_summary = (
        "Use these settings to control Orca's caret navigation behavior in document content."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for caret navigation preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.KB_GROUP_CARET_NAVIGATION,
            panel_id="document_presenter.caret-navigation",
            description=(
                "Caret navigation settings apply when moving through document content "
                "with caret commands."
            ),
            schema="caret-navigation",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.KB_GROUP_CARET_NAVIGATION,
                    kind="switch",
                    summary="Controls whether Orca handles caret navigation commands.",
                    schema="caret-navigation",
                    key=caret_navigator.CaretNavigator.KEY_ENABLED,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.AUTOMATIC_FOCUS_MODE,
                    kind="switch",
                    summary=(
                        "Controls whether caret navigation can cause Orca to switch to focus mode."
                    ),
                    schema="caret-navigation",
                    key=caret_navigator.CaretNavigator.KEY_TRIGGERS_FOCUS_MODE,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.CONTENT_LAYOUT_MODE,
                    kind="switch",
                    summary=(
                        "Controls whether Orca presents each full visual line or treats "
                        "each object as if it were on a separate line."
                    ),
                    schema="caret-navigation",
                    key=caret_navigator.CaretNavigator.KEY_LAYOUT_MODE,
                ),
            ),
        )

    def __init__(self) -> None:
        nav = caret_navigator.get_navigator()

        # Child controls need to check the enabled switch's UI state (not runtime state)
        # because the enabled switch has apply_immediately=False.
        self._enabled_switch: Gtk.Switch | None = None

        def is_enabled() -> bool:
            if self._enabled_switch is not None:
                return self._enabled_switch.get_active()
            return nav.get_is_enabled()

        controls = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.KB_GROUP_CARET_NAVIGATION,
                getter=nav.get_is_enabled,
                setter=nav.set_is_enabled,
                prefs_key=caret_navigator.CaretNavigator.KEY_ENABLED,
                apply_immediately=False,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.AUTOMATIC_FOCUS_MODE,
                getter=nav.get_triggers_focus_mode,
                setter=nav.set_triggers_focus_mode,
                prefs_key=caret_navigator.CaretNavigator.KEY_TRIGGERS_FOCUS_MODE,
                determine_sensitivity=is_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.CONTENT_LAYOUT_MODE,
                getter=nav.get_layout_mode,
                setter=nav.set_layout_mode,
                prefs_key=caret_navigator.CaretNavigator.KEY_LAYOUT_MODE,
                determine_sensitivity=is_enabled,
            ),
        ]
        info = (
            f"{guilabels.CARET_NAVIGATION_INFO}\n\n{guilabels.AUTOMATIC_FOCUS_MODE_INFO}"
            f"\n\n{guilabels.LAYOUT_MODE_INFO}"
        )
        super().__init__(guilabels.KB_GROUP_CARET_NAVIGATION, controls, info_message=info)

        self._enabled_switch = self._widgets[0]


class StructuralNavigationPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """Sub-grid for structural navigation settings within the Documents page."""

    _gsettings_schema = "structural-navigation"
    _documentation_summary = (
        "Use these settings to control navigation by headings, landmarks, form fields, "
        "and other document structures."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for structural navigation preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.KB_GROUP_STRUCTURAL_NAVIGATION,
            panel_id="document_presenter.structural-navigation",
            description=(
                "Structural navigation settings apply when moving among document "
                "elements such as headings, links, and form fields."
            ),
            schema="structural-navigation",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.KB_GROUP_STRUCTURAL_NAVIGATION,
                    kind="switch",
                    summary=(
                        "Controls whether Orca's structural navigation commands are available."
                    ),
                    schema="structural-navigation",
                    key=structural_navigator.StructuralNavigator.KEY_ENABLED,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.AUTOMATIC_FOCUS_MODE,
                    kind="switch",
                    summary=(
                        "Controls whether structural navigation can cause Orca to switch "
                        "to focus mode."
                    ),
                    schema="structural-navigation",
                    key=structural_navigator.StructuralNavigator.KEY_TRIGGERS_FOCUS_MODE,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.STRUCTURAL_NAVIGATION_WRAP_AROUND,
                    kind="switch",
                    summary=(
                        "Controls whether structural navigation wraps to the other end "
                        "of the document."
                    ),
                    schema="structural-navigation",
                    key=structural_navigator.StructuralNavigator.KEY_WRAPS,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.STRUCTURAL_NAVIGATION_LARGE_OBJECT_LENGTH,
                    kind="spin",
                    summary=(
                        "Sets the minimum number of characters for an element to be "
                        "included by large object navigation commands."
                    ),
                    schema="structural-navigation",
                    key=structural_navigator.StructuralNavigator.KEY_LARGE_OBJECT_TEXT_LENGTH,
                    minimum="1",
                    maximum="500",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.STRUCTURAL_NAVIGATION_SKIP_UNLABELED_IMAGES,
                    kind="switch",
                    summary=(
                        "Controls whether image navigation skips images without accessible "
                        "names or descriptions."
                    ),
                    schema="structural-navigation",
                    key=structural_navigator.StructuralNavigator.KEY_SKIP_UNLABELED_IMAGES,
                ),
            ),
        )

    def __init__(self) -> None:
        nav = structural_navigator.get_navigator()

        # Child controls need to check the enabled switch's UI state (not runtime state)
        # because the enabled switch has apply_immediately=False.
        self._enabled_switch: Gtk.Switch | None = None

        def is_enabled() -> bool:
            if self._enabled_switch is not None:
                return self._enabled_switch.get_active()
            return nav.get_is_enabled()

        controls: list[preferences_grid_base.ControlType] = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.KB_GROUP_STRUCTURAL_NAVIGATION,
                getter=nav.get_is_enabled,
                setter=nav.set_is_enabled,
                prefs_key=structural_navigator.StructuralNavigator.KEY_ENABLED,
                apply_immediately=False,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.AUTOMATIC_FOCUS_MODE,
                getter=nav.get_triggers_focus_mode,
                setter=nav.set_triggers_focus_mode,
                prefs_key=structural_navigator.StructuralNavigator.KEY_TRIGGERS_FOCUS_MODE,
                determine_sensitivity=is_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.STRUCTURAL_NAVIGATION_WRAP_AROUND,
                getter=nav.get_navigation_wraps,
                setter=nav.set_navigation_wraps,
                prefs_key=structural_navigator.StructuralNavigator.KEY_WRAPS,
                determine_sensitivity=is_enabled,
            ),
            preferences_grid_base.IntRangePreferenceControl(
                label=guilabels.STRUCTURAL_NAVIGATION_LARGE_OBJECT_LENGTH,
                minimum=1,
                maximum=500,
                getter=nav.get_large_object_text_length,
                setter=nav.set_large_object_text_length,
                prefs_key=structural_navigator.StructuralNavigator.KEY_LARGE_OBJECT_TEXT_LENGTH,
                determine_sensitivity=is_enabled,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.STRUCTURAL_NAVIGATION_SKIP_UNLABELED_IMAGES,
                getter=nav.get_skip_unlabeled_images,
                setter=nav.set_skip_unlabeled_images,
                prefs_key=structural_navigator.StructuralNavigator.KEY_SKIP_UNLABELED_IMAGES,
                determine_sensitivity=is_enabled,
            ),
        ]
        info = (
            f"{guilabels.STRUCTURAL_NAVIGATION_INFO}\n\n{guilabels.AUTOMATIC_FOCUS_MODE_INFO}"
            f"\n\n{guilabels.LARGE_OBJECT_INFO}"
        )
        super().__init__(guilabels.KB_GROUP_STRUCTURAL_NAVIGATION, controls, info_message=info)

        self._enabled_switch = self._widgets[0]


class TableNavigationPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """Sub-grid for table navigation settings within the Documents page."""

    _gsettings_schema = "table-navigation"
    _documentation_summary = "Use these settings to control table navigation behavior."

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for table navigation preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.KB_GROUP_TABLE_NAVIGATION,
            panel_id="document_presenter.table-navigation",
            description=(
                "Table navigation lets you use Orca commands to move by table cell, "
                "row, and column in document content."
            ),
            schema="table-navigation",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.KB_GROUP_TABLE_NAVIGATION,
                    kind="switch",
                    summary=("Controls whether Orca's table navigation commands are available."),
                    schema="table-navigation",
                    key=table_navigator.TableNavigator.KEY_ENABLED,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.TABLE_SKIP_BLANK_CELLS,
                    kind="switch",
                    summary=("Controls whether table navigation skips empty cells."),
                    schema="table-navigation",
                    key=table_navigator.TableNavigator.KEY_SKIP_BLANK_CELLS,
                ),
            ),
        )

    def __init__(self) -> None:
        nav = table_navigator.get_navigator()

        # Child controls need to check the enabled switch's UI state (not runtime state)
        # because the enabled switch has apply_immediately=False.
        self._enabled_switch: Gtk.Switch | None = None

        def is_enabled() -> bool:
            if self._enabled_switch is not None:
                return self._enabled_switch.get_active()
            return nav.get_is_enabled()

        controls = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.KB_GROUP_TABLE_NAVIGATION,
                getter=nav.get_is_enabled,
                setter=nav.set_is_enabled,
                prefs_key=table_navigator.TableNavigator.KEY_ENABLED,
                apply_immediately=False,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.TABLE_SKIP_BLANK_CELLS,
                getter=nav.get_skip_blank_cells,
                setter=nav.set_skip_blank_cells,
                prefs_key=table_navigator.TableNavigator.KEY_SKIP_BLANK_CELLS,
                determine_sensitivity=is_enabled,
            ),
        ]
        super().__init__(guilabels.KB_GROUP_TABLE_NAVIGATION, controls)

        self._enabled_switch = self._widgets[0]


class NativeNavigationPreferencesGrid(preferences_grid_base.AutoPreferencesGrid):
    """Sub-grid for native navigation settings within the Documents page."""

    _gsettings_schema = "document"
    _documentation_summary = (
        "Use these settings to control how Orca responds to application-provided navigation "
        "and find results."
    )

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for native navigation preferences."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.NATIVE_NAVIGATION,
            panel_id="document_presenter.native-navigation",
            description=(
                "Native navigation settings apply to application-handled navigation, "
                "find results, and page loading."
            ),
            schema="document",
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.AUTOMATIC_FOCUS_MODE,
                    kind="switch",
                    summary=(
                        "Controls whether native navigation can cause Orca to switch to focus mode."
                    ),
                    schema="document",
                    key=document_presenter.DocumentPresenter.KEY_NATIVE_NAV_TRIGGERS_FOCUS_MODE,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.AUTO_STICKY_FOCUS_MODE,
                    kind="switch",
                    summary=(
                        "Controls whether Orca detects web applications and Electron apps "
                        "and stays in focus mode."
                    ),
                    schema="document",
                    key=document_presenter.DocumentPresenter.KEY_AUTO_STICKY_FOCUS_MODE,
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.FIND_OPTIONS,
                    kind="group",
                    summary=(
                        "Controls what Orca speaks while you use an application's Find toolbar."
                    ),
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.FIND_SPEAK_RESULTS,
                            kind="switch",
                            summary=(
                                "Controls whether Orca speaks the line containing each find result."
                            ),
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.FIND_ONLY_SPEAK_CHANGED_LINES,
                            kind="switch",
                            summary=(
                                "Controls whether Orca speaks a find result only when it is "
                                "on a different line than the previous result."
                            ),
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.FIND_MINIMUM_MATCH_LENGTH,
                            kind="spin",
                            summary=(
                                "Sets the minimum number of matching characters required "
                                "before Orca speaks the find result."
                            ),
                            schema="document",
                            key=(
                                document_presenter.DocumentPresenter.KEY_FIND_RESULTS_MINIMUM_LENGTH
                            ),
                            minimum="0",
                            maximum="20",
                        ),
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.PAGE_LOAD,
                    kind="group",
                    summary=(
                        "Controls what Orca does when a document or web page finishes loading."
                    ),
                    controls=(
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.READ_PAGE_UPON_LOAD,
                            kind="switch",
                            summary=(
                                "Controls whether Orca automatically performs Say All when "
                                "a document or web page first loads."
                            ),
                            schema="document",
                            key=document_presenter.DocumentPresenter.KEY_SAY_ALL_ON_LOAD,
                        ),
                        preferences_grid_base.PreferenceControlDoc(
                            label=guilabels.PAGE_SUMMARY_UPON_LOAD,
                            kind="switch",
                            summary=(
                                "Controls whether Orca presents a summary, such as the "
                                "number of headings, landmarks, and links, when a document "
                                "or web page first loads."
                            ),
                            schema="document",
                            key=document_presenter.DocumentPresenter.KEY_PAGE_SUMMARY_ON_LOAD,
                        ),
                    ),
                ),
            ),
        )

    def __init__(self, presenter: DocumentPresenter) -> None:
        self._presenter = presenter
        controls: list[preferences_grid_base.ControlType] = [
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.AUTOMATIC_FOCUS_MODE,
                getter=presenter.get_native_nav_triggers_focus_mode,
                setter=presenter.set_native_nav_triggers_focus_mode,
                prefs_key=document_presenter.DocumentPresenter.KEY_NATIVE_NAV_TRIGGERS_FOCUS_MODE,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.AUTO_STICKY_FOCUS_MODE,
                getter=presenter.get_auto_sticky_focus_mode_for_web_apps,
                setter=presenter.set_auto_sticky_focus_mode_for_web_apps,
                prefs_key=document_presenter.DocumentPresenter.KEY_AUTO_STICKY_FOCUS_MODE,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.FIND_SPEAK_RESULTS,
                getter=presenter.get_speak_find_results,
                setter=presenter.set_speak_find_results,
                member_of=guilabels.FIND_OPTIONS,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.FIND_ONLY_SPEAK_CHANGED_LINES,
                getter=presenter.get_only_speak_changed_lines,
                setter=presenter.set_only_speak_changed_lines,
                determine_sensitivity=presenter.get_speak_find_results,
                member_of=guilabels.FIND_OPTIONS,
            ),
            preferences_grid_base.IntRangePreferenceControl(
                label=guilabels.FIND_MINIMUM_MATCH_LENGTH,
                minimum=0,
                maximum=20,
                getter=presenter.get_find_results_minimum_length,
                setter=presenter.set_find_results_minimum_length,
                prefs_key=document_presenter.DocumentPresenter.KEY_FIND_RESULTS_MINIMUM_LENGTH,
                member_of=guilabels.FIND_OPTIONS,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.READ_PAGE_UPON_LOAD,
                getter=presenter.get_say_all_on_load,
                setter=presenter.set_say_all_on_load,
                prefs_key=document_presenter.DocumentPresenter.KEY_SAY_ALL_ON_LOAD,
                member_of=guilabels.PAGE_LOAD,
            ),
            preferences_grid_base.BooleanPreferenceControl(
                label=guilabels.PAGE_SUMMARY_UPON_LOAD,
                getter=presenter.get_page_summary_on_load,
                setter=presenter.set_page_summary_on_load,
                prefs_key=document_presenter.DocumentPresenter.KEY_PAGE_SUMMARY_ON_LOAD,
                member_of=guilabels.PAGE_LOAD,
            ),
        ]
        info = (
            f"{guilabels.NATIVE_NAVIGATION_INFO}\n\n"
            f"{guilabels.AUTOMATIC_FOCUS_MODE_INFO}\n\n"
            f"{guilabels.AUTO_STICKY_FOCUS_MODE_INFO}"
        )
        super().__init__(guilabels.NATIVE_NAVIGATION, controls, info_message=info)

    def save_settings(self, profile: str = "", app_name: str = "") -> dict[str, Any]:
        """Save settings, writing the find-results enum from the presenter."""

        result = super().save_settings(profile, app_name)
        verbosity = self._presenter.get_find_results_verbosity_name()
        result[document_presenter.DocumentPresenter.KEY_FIND_RESULTS_VERBOSITY] = verbosity
        self._write_gsettings(
            {document_presenter.DocumentPresenter.KEY_FIND_RESULTS_VERBOSITY: verbosity},
            profile,
            app_name,
        )
        return result


class DocumentPreferencesGrid(preferences_grid_base.PreferencesGridBase):
    """Main document preferences grid with categorized navigation settings."""

    @classmethod
    def get_documentation(cls) -> preferences_grid_base.PreferencePanelDoc:
        """Return documentation metadata for the documents preferences landing page."""

        return preferences_grid_base.PreferencePanelDoc(
            title=guilabels.DOCUMENTS,
            panel_id="manual.documents",
            description=(
                "Document settings control how Orca navigates and presents documents, "
                "web pages, and math content.\n\n"
                "Activate a row, or press Right Arrow, to open its settings. Press "
                "Left Arrow, Escape, or Alt+Left to return to the list."
            ),
            show_available_settings=False,
            controls=(
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.KB_GROUP_CARET_NAVIGATION,
                    kind="page",
                    summary="Opens settings for caret navigation in document content.",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.KB_GROUP_STRUCTURAL_NAVIGATION,
                    kind="page",
                    summary=(
                        "Opens settings for navigation by headings, landmarks, links, "
                        "and other structures."
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.KB_GROUP_TABLE_NAVIGATION,
                    kind="page",
                    summary="Opens settings for table navigation in document content.",
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.NATIVE_NAVIGATION,
                    kind="page",
                    summary=(
                        "Opens settings for application-provided navigation, find results, "
                        "and page loading."
                    ),
                ),
                preferences_grid_base.PreferenceControlDoc(
                    label=guilabels.MATH_PRESENTATION,
                    kind="page",
                    summary="Opens settings for presenting and navigating math content.",
                ),
            ),
        )

    def __init__(
        self,
        presenter: DocumentPresenter,
        title_change_callback: Callable[[str], None] | None = None,
    ) -> None:
        super().__init__(guilabels.DOCUMENTS)
        self._presenter = presenter
        self._initializing = True
        self._title_change_callback = title_change_callback

        self._caret_grid = CaretNavigationPreferencesGrid()
        self._structural_grid = StructuralNavigationPreferencesGrid()
        self._table_grid = TableNavigationPreferencesGrid()
        self._native_grid = NativeNavigationPreferencesGrid(presenter)
        self._math_grid = math_presenter.get_presenter().create_preferences_grid()

        self._build()
        self._initializing = False

    def _build(self) -> None:
        categories = [
            (guilabels.KB_GROUP_CARET_NAVIGATION, "caret", self._caret_grid),
            (guilabels.KB_GROUP_STRUCTURAL_NAVIGATION, "structural", self._structural_grid),
            (guilabels.KB_GROUP_TABLE_NAVIGATION, "table", self._table_grid),
            (guilabels.NATIVE_NAVIGATION, "native", self._native_grid),
            (guilabels.MATH_PRESENTATION, "math", self._math_grid),
        ]

        if not math_presenter.get_presenter().is_available():
            categories.remove((guilabels.MATH_PRESENTATION, "math", self._math_grid))

        _enable_listbox, stack, _categories_listbox = self._create_child_grid_preferences_stack(
            enable_label=None,
            enable_getter=None,
            enable_setter=None,
            categories=categories,
            title_change_callback=self._title_change_callback,
            main_title=guilabels.DOCUMENTS,
        )

        self.attach(stack, 0, 0, 1, 1)

    def on_becoming_visible(self) -> None:
        """Reset to the categories view when this grid becomes visible."""

        self.multipage_on_becoming_visible()

    def reload(self) -> None:
        """Reload all child grids."""

        self._initializing = True
        self._has_unsaved_changes = False
        self._caret_grid.reload()
        self._structural_grid.reload()
        self._table_grid.reload()
        self._native_grid.reload()
        self._math_grid.reload()
        self._initializing = False

    def save_settings(self, profile: str = "", app_name: str = "") -> dict:
        """Save all settings from child grids."""

        result = {}
        result.update(self._caret_grid.save_settings(profile, app_name))
        result.update(self._structural_grid.save_settings(profile, app_name))
        result.update(self._table_grid.save_settings(profile, app_name))
        result.update(self._native_grid.save_settings(profile, app_name))
        result.update(self._math_grid.save_settings(profile, app_name))

        return result

    def has_changes(self) -> bool:
        """Check if any child grid has changes."""

        return (
            self._has_unsaved_changes
            or self._caret_grid.has_changes()
            or self._structural_grid.has_changes()
            or self._table_grid.has_changes()
            or self._native_grid.has_changes()
            or self._math_grid.has_changes()
        )

    def refresh(self) -> None:
        """Refresh all child grids."""

        self._initializing = True
        self._caret_grid.refresh()
        self._structural_grid.refresh()
        self._table_grid.refresh()
        self._native_grid.refresh()
        self._math_grid.refresh()
        self._initializing = False



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2011-2026 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-public-methods
# pylint: disable=too-many-lines

"""Provides braille presentation support."""

from __future__ import annotations

import os
import time
from enum import Enum
from typing import TYPE_CHECKING, Any

from . import (
    braille,
    braille_monitor,
    braille_presenter_command_definitions,
    brltablenames,
    clipboard,
    dbus_service,
    debug,
    document_presenter,
    extension_loader,
    flat_review_presenter,
    focus_manager,
    gsettings_registry,
    guilabels,
    input_event,
    input_event_manager,
    messages,
    output_recorder,
    presentation_manager,
)
from .ax_text import AXText
from .ax_utilities import AXUtilities
from .ax_utilities_text import CaretSetReason
from .braille_generator import BrailleGeneratorContext
from .extension import BrailleOutput, BrailleOutputResult, Extension
from .generator import PresentationReason
from .orca_platform import tablesdir  # pylint: disable=import-error

if TYPE_CHECKING:
    from collections.abc import Callable

    import gi

    from .dbus_service import UInt32

    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .braille_presenter_preferences_grid import BraillePreferencesGrid
    from .command import Command
    from .scripts import default


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.VerbosityLevel",
    values={"brief": 0, "verbose": 1},
)
class VerbosityLevel(Enum):
    """Verbosity level enumeration."""

    BRIEF = 0
    VERBOSE = 1

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower()


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.BrailleIndicator",
    values={"none": 0, "dot7": 64, "dot8": 128, "dots78": 192},
)
class BrailleIndicator(Enum):
    """Braille indicator enumeration."""

    NONE = 0x00
    DOT7 = 0x40
    DOT8 = 0x80
    DOTS78 = 0xC0

    @property
    def string_name(self) -> str:
        """Returns the lowercase string name for this enum value."""

        return self.name.lower()


@gsettings_registry.get_registry().gsettings_enum(
    "org.gnome.Orca.ProgressBarVerbosity",
    values={"all": 0, "application": 1, "window": 2},
)
class ProgressBarVerbosity(Enum):
    """Progress bar verbosity level enumeration."""

    ALL = 0
    APPLICATION = 1
    WINDOW = 2


class PanDirection(Enum):
    """Braille panning direction."""

    LEFT = "left"
    RIGHT = "right"


@gsettings_registry.get_registry().gsettings_schema("org.gnome.Orca.Braille", name="braille")
class BraillePresenter(Extension):
    """Provides braille presentation support."""

    GROUP_LABEL = guilabels.BRAILLE

    _SCHEMA = "braille"

    KEY_ENABLED = "enabled"
    KEY_VERBOSITY_LEVEL = "verbosity-level"
    KEY_ROLENAME_STYLE = "rolename-style"
    KEY_PRESENT_MNEMONICS = "present-mnemonics"
    KEY_DISPLAY_ANCESTORS = "display-ancestors"
    KEY_BRAILLE_PROGRESS_BAR_UPDATES = "braille-progress-bar-updates"
    KEY_PROGRESS_BAR_BRAILLE_INTERVAL = "progress-bar-braille-interval"
    KEY_PROGRESS_BAR_BRAILLE_VERBOSITY = "progress-bar-braille-verbosity"
    KEY_CONTRACTED_BRAILLE = "contracted-braille"
    KEY_COMPUTER_BRAILLE_AT_CURSOR = "computer-braille-at-cursor"
    KEY_CONTRACTION_TABLE = "contraction-table"
    KEY_END_OF_LINE_INDICATOR = "end-of-line-indicator"
    KEY_WORD_WRAP = "word-wrap"
    KEY_FLASH_MESSAGES = "flash-messages"
    KEY_FLASH_MESSAGE_DURATION = "flash-message-duration"
    KEY_FLASH_MESSAGES_PERSISTENT = "flash-messages-persistent"
    KEY_FLASH_MESSAGES_DETAILED = "flash-messages-detailed"
    KEY_SELECTOR_INDICATOR = "selector-indicator"
    KEY_LINK_INDICATOR = "link-indicator"
    KEY_TEXT_ATTRIBUTES_INDICATOR = "text-attributes-indicator"
    KEY_MONITOR_CELL_COUNT = "monitor-cell-count"
    KEY_MONITOR_SHOW_DOTS = "monitor-show-dots"
    KEY_MONITOR_FOREGROUND = "monitor-foreground"
    KEY_MONITOR_BACKGROUND = "monitor-background"

    def _get_setting(self, key: str, gtype: str, default: Any) -> Any:
        """Returns the dconf value for key, or default if not in dconf."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            key,
            gtype,
            default=default,
        )

    def __init__(self) -> None:
        self._command_names: dict[int, str] | None = None
        self._monitor: braille_monitor.BrailleMonitor | None = None
        self._monitor_enabled_override: bool | None = None
        self._progress_bar_cache: dict = {}
        self._output_recorder = output_recorder.OutputRecorder("braille")
        super().__init__()

    def _get_commands(self) -> list[Command]:
        """Returns commands for registration."""

        return braille_presenter_command_definitions.get_commands(self)

    def pan_braille_left(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Pans the braille display to the left."""

        return self._pan_braille(PanDirection.LEFT, script, event)

    def pan_braille_right(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Pans the braille display to the right."""

        return self._pan_braille(PanDirection.RIGHT, script, event)

    def _pan_braille(
        self,
        direction: PanDirection,
        script: default.Script | None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Pans braille in direction, asking the script only when at an edge."""

        if isinstance(event, input_event.KeyboardEvent) and not self.use_braille():
            msg = f"BRAILLE PRESENTER: panBraille{direction.name.title()} command requires braille"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            return True

        flat_review_result = self._pan_flat_review_braille(direction, script, event)
        if flat_review_result is not None:
            return flat_review_result

        did_pan = self.pan_left() if direction == PanDirection.LEFT else self.pan_right()
        if did_pan:
            return True

        if script is not None:
            script_result = script.handle_braille_pan_at_edge(direction)
            if script_result is not None:
                return script_result

        return self._handle_braille_pan_at_edge(direction, script, event)

    @staticmethod
    def _pan_flat_review_braille(
        direction: PanDirection,
        script: default.Script | None,
        event: input_event.InputEvent | None = None,
    ) -> bool | None:
        """Pans flat review braille if flat review is active."""

        flat_review = flat_review_presenter.get_presenter()
        if not flat_review.is_active():
            return None
        if script is None:
            return True
        if direction == PanDirection.LEFT:
            return flat_review.pan_braille_left(script, event)
        return flat_review.pan_braille_right(script, event)

    def _handle_braille_pan_at_edge(
        self,
        direction: PanDirection,
        script: default.Script | None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Handles braille panning when no more cells are available in the current line."""

        focus = focus_manager.get_manager().get_locus_of_focus()
        is_text_area = AXUtilities.is_editable(focus) or AXUtilities.is_terminal(focus)
        if not is_text_area:
            return True

        if direction == PanDirection.LEFT:
            start_offset = AXText.get_line_at_offset(focus)[1]
            moved_caret = False
            if start_offset > 0:
                moved_caret = AXUtilities.set_caret_offset_with_reason(
                    focus, start_offset - 1, CaretSetReason.BRAILLE_PANNING
                )

            # If we didn't move the caret and we're in a terminal, jump into flat review to
            # review the text. See http://bugzilla.gnome.org/show_bug.cgi?id=482294.
            if not moved_caret and script is not None and AXUtilities.is_terminal(focus):
                flat_review = flat_review_presenter.get_presenter()
                flat_review.go_start_of_line(script, event)
                flat_review.go_previous_character(script, event)
            return True

        end_offset = AXText.get_line_at_offset(focus)[2]
        if end_offset < AXText.get_character_count(focus):
            AXUtilities.set_caret_offset_with_reason(
                focus, end_offset, CaretSetReason.BRAILLE_PANNING
            )

        return True

    def go_home(
        self,
        _script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Returns to the component with focus."""

        if flat_review_presenter.get_presenter().is_active():
            flat_review_presenter.get_presenter().quit()
            return True

        presentation_manager.get_manager().interrupt_presentation()
        return braille.return_to_region_with_focus(event)

    def toggle_contracted_braille(
        self,
        _script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
    ) -> bool:
        """Toggles contracted braille."""

        braille.toggle_contracted_braille(event)
        return True

    def process_routing_key(
        self,
        _script: default.Script | None = None,
        event: input_event.BrailleEvent | None = None,
    ) -> bool:
        """Processes a cursor routing key."""

        # Don't kill flash here because it will restore the previous contents and
        # then process the routing key. If the contents accept a click action, this
        # would result in clicking on the link instead of clearing the flash message.
        presentation_manager.get_manager().interrupt_presentation(kill_flash=False)
        if event is None:
            return True
        braille.process_routing_key(event)
        return True

    def process_braille_cut_begin(
        self,
        script: default.Script | None = None,
        event: input_event.BrailleEvent | None = None,
    ) -> bool:
        """Clears the selection and moves the caret offset in the current text area."""

        if event is None:
            return True
        caret_context = braille.get_caret_context(event)
        if caret_context.offset < 0:
            return True

        presentation_manager.get_manager().interrupt_presentation()
        AXUtilities.clear_all_selected_text(caret_context.accessible)
        if script is not None:
            script.utilities.set_caret_offset(
                caret_context.accessible,
                caret_context.offset,
                reason=CaretSetReason.BRAILLE_CUT,
            )
        else:
            AXUtilities.set_caret_offset_with_reason(
                caret_context.accessible,
                caret_context.offset,
                CaretSetReason.BRAILLE_CUT,
            )
        return True

    def process_braille_cut_line(
        self,
        _script: default.Script | None = None,
        event: input_event.BrailleEvent | None = None,
    ) -> bool:
        """Extends the current text selection and copies it to the clipboard."""

        if event is None:
            return True
        caret_context = braille.get_caret_context(event)
        if caret_context.offset < 0:
            return True

        presentation_manager.get_manager().interrupt_presentation()
        start_offset = AXUtilities.get_selection_start_offset(caret_context.accessible)
        end_offset = AXUtilities.get_selection_end_offset(caret_context.accessible)
        if start_offset < 0 or end_offset < 0:
            caret_offset = AXText.get_caret_offset(caret_context.accessible)
            start_offset = min(caret_context.offset, caret_offset)
            end_offset = max(caret_context.offset, caret_offset)

        AXUtilities.set_selected_text(caret_context.accessible, start_offset, end_offset)
        text = AXUtilities.get_selected_text(caret_context.accessible)[0]
        clipboard.get_presenter().set_text(text)
        return True

    @dbus_service.command
    def toggle_monitor(
        self,
        script: default.Script | None = None,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Toggles the braille monitor on and off."""

        tokens = [
            "BRAILLE PRESENTER: toggle_monitor. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if self.get_monitor_is_enabled():
            self.set_monitor_is_enabled(False)
            if script is not None and notify_user:
                presentation_manager.get_manager().present_message(
                    messages.BRAILLE_MONITOR_DISABLED,
                )
        else:
            self.set_monitor_is_enabled(True)
            if script is not None and notify_user:
                presentation_manager.get_manager().present_message(messages.BRAILLE_MONITOR_ENABLED)
        return True

    def get_table_names(self) -> dict[str, str]:
        """Returns table filenames mapped to localized display names."""

        return dict(brltablenames.TABLE_NAMES)

    @staticmethod
    def _build_command_names() -> dict[int, str]:
        """Return BrlTTY command names for presentation in the UI."""

        return braille_presenter_command_definitions.get_brltty_command_names()

    def get_command_names(self) -> dict[int, str]:
        """Returns a mapping of BrlTTY command IDs to user-visible labels."""

        if self._command_names is None:
            self._command_names = self._build_command_names()
        return dict(self._command_names)

    def create_preferences_grid(
        self,
        title_change_callback: Callable[[str], None] | None = None,
    ) -> BraillePreferencesGrid:
        """Returns the GtkGrid containing the preferences UI."""

        # pylint: disable-next=import-outside-toplevel
        from .braille_presenter_preferences_grid import BraillePreferencesGrid

        return BraillePreferencesGrid(self, title_change_callback)

    def _get_verbosity_is_detailed(self) -> bool:
        """Returns whether braille verbosity is set to detailed/verbose."""

        return self.get_verbosity_level() == "verbose"

    def _set_verbosity_is_detailed(self, value: bool) -> bool:
        """Sets braille verbosity to detailed/verbose if True, brief if False."""

        return self.set_verbosity_level("verbose" if value else "brief")

    def _get_use_abbreviated_rolenames(self) -> bool:
        """Returns whether abbreviated role names are used."""

        return self.get_rolename_style() == "brief"

    def _set_use_abbreviated_rolenames(self, value: bool) -> bool:
        """Sets whether abbreviated role names are used."""

        return self.set_rolename_style("brief" if value else "verbose")

    def _get_flash_duration_seconds(self) -> int:
        """Returns flash duration in seconds (converted from milliseconds)."""

        duration_ms = self.get_flash_message_duration()
        return max(1, duration_ms // 1000)

    def _set_flash_duration_seconds(self, value: int) -> None:
        """Sets flash duration in seconds (converts to milliseconds)."""

        duration_ms = value * 1000
        self.set_flash_message_duration(duration_ms)

    def use_braille(self) -> bool:
        """Returns whether braille is to be used."""

        result = self.get_braille_is_enabled() or self.get_monitor_is_enabled()
        if not result:
            msg = "BRAILLE PRESENTER: Braille is disabled."
            debug.print_message(debug.LEVEL_INFO, msg, True)
        return result

    @dbus_service.getter
    def get_monitor_is_enabled(self) -> bool:
        """Returns whether the braille monitor is enabled."""

        return self._monitor_enabled_override or False

    @dbus_service.setter
    def set_monitor_is_enabled(self, value: bool) -> bool:
        """Sets whether the braille monitor is enabled."""

        msg = f"BRAILLE PRESENTER: Setting enable braille monitor to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        self._monitor_enabled_override = value
        if value:
            braille.set_monitor_cell_count(self.get_monitor_cell_count())
        else:
            braille.set_monitor_cell_count(0)
            self.destroy_monitor()
        return True

    @dbus_service.testing_command
    def set_log_file_for_testing(
        self,
        token: str = "",  # pylint: disable=unused-argument
        value: str = "",
        script: default.Script | None = None,  # pylint: disable=unused-argument
        event: input_event.InputEvent | None = None,  # pylint: disable=unused-argument
    ) -> bool:
        """Opens value for JSONL recording; an empty string closes any open file (test-only)."""

        msg = f"BRAILLE PRESENTER: Setting log file to {value!r}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return self._output_recorder.set_path(value)

    @gsettings_registry.get_registry().gsetting(
        key=KEY_MONITOR_CELL_COUNT,
        schema="braille",
        gtype="i",
        default=32,
        summary="Braille monitor cell count",
        migration_key="brailleMonitorCellCount",
    )
    @dbus_service.getter
    def get_monitor_cell_count(self) -> UInt32:
        """Returns the configured braille monitor cell count."""

        return self._get_setting(self.KEY_MONITOR_CELL_COUNT, "i", 32)

    @dbus_service.setter
    def set_monitor_cell_count(self, value: UInt32) -> bool:
        """Sets the braille monitor cell count."""

        msg = f"BRAILLE PRESENTER: Setting braille monitor cell count to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_MONITOR_CELL_COUNT,
            value,
        )
        if self.get_monitor_is_enabled():
            braille.set_monitor_cell_count(value)
        self.destroy_monitor()
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_MONITOR_SHOW_DOTS,
        schema="braille",
        gtype="b",
        default=False,
        summary="Show Unicode braille dots in braille monitor",
        migration_key="brailleMonitorShowDots",
    )
    @dbus_service.getter
    def get_monitor_show_dots(self) -> bool:
        """Returns whether the braille monitor shows Unicode braille dots."""

        return self._get_setting(self.KEY_MONITOR_SHOW_DOTS, "b", False)

    @dbus_service.setter
    def set_monitor_show_dots(self, value: bool) -> bool:
        """Sets whether the braille monitor shows Unicode braille dots."""

        msg = f"BRAILLE PRESENTER: Setting braille monitor show dots to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_MONITOR_SHOW_DOTS,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_MONITOR_FOREGROUND,
        schema="braille",
        gtype="s",
        default="#000000",
        summary="Braille monitor foreground color",
        migration_key="brailleMonitorForeground",
    )
    @dbus_service.getter
    def get_monitor_foreground(self) -> str:
        """Returns the braille monitor foreground color."""

        return self._get_setting(self.KEY_MONITOR_FOREGROUND, "s", "#000000")

    @dbus_service.setter
    def set_monitor_foreground(self, value: str) -> bool:
        """Sets the braille monitor foreground color."""

        msg = f"BRAILLE PRESENTER: Setting braille monitor foreground to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_MONITOR_FOREGROUND,
            value,
        )
        if self._monitor is not None:
            self._monitor.reapply_css(foreground=value)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_MONITOR_BACKGROUND,
        schema="braille",
        gtype="s",
        default="#ffffff",
        summary="Braille monitor background color",
        migration_key="brailleMonitorBackground",
    )
    @dbus_service.getter
    def get_monitor_background(self) -> str:
        """Returns the braille monitor background color."""

        return self._get_setting(self.KEY_MONITOR_BACKGROUND, "s", "#ffffff")

    @dbus_service.setter
    def set_monitor_background(self, value: str) -> bool:
        """Sets the braille monitor background color."""

        msg = f"BRAILLE PRESENTER: Setting braille monitor background to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_MONITOR_BACKGROUND,
            value,
        )
        if self._monitor is not None:
            self._monitor.reapply_css(background=value)
        return True

    def present_regions(
        self,
        regions: list[braille.Region],
        focused_region: braille.Region | None,
        *,
        pan_to_cursor: bool = True,
        stop_flash: bool = True,
    ) -> None:
        """Build a line from regions and present it as a single braille line."""

        regions, focused_region, consumed = self._process_braille_output(
            regions,
            focused_region,
        )
        if consumed:
            return

        line = braille.Line()
        line.add_regions(regions)
        braille.display_line(
            line,
            focused_region,
            pan_to_cursor=pan_to_cursor,
            stop_flash=stop_flash,
        )

    def _build_generator_context(
        self,
        reason: PresentationReason | None = None,
        prior_obj: Atspi.Accessible | None = None,
        offset: int | None = None,
    ) -> BrailleGeneratorContext:
        """Builds the settings context for braille generators."""

        mgr = focus_manager.get_manager()
        active_mode, _obj = mgr.get_active_mode_and_object_of_interest()

        return BrailleGeneratorContext(
            enabled=self.use_braille(),
            verbose=self.use_verbose_braille(),
            focus=mgr.get_locus_of_focus(),
            in_focus_mode=document_presenter.get_presenter().get_in_focus_mode(),
            active_mode=active_mode,
            reason=reason or PresentationReason.FOCUS_CHANGE,
            prior_obj=prior_obj,
            offset=offset,
            leaving=False,
            ancestor_of=None,
            content_item=None,
            content_position=None,
            content_subject=None,
            resolved_role=None,
            role_subject=None,
            include_context=True,
            full_rolenames=self.use_full_rolenames(),
            display_ancestors=self.get_display_ancestors(),
            end_of_line_indicator=self.get_end_of_line_indicator_is_enabled(),
            present_mnemonics=self.get_present_mnemonics(),
            indicate_links=True,
        )

    def display_generated_contents(
        self,
        script: default.Script,
        contents: list[tuple[Atspi.Accessible, int, int, str]],
    ) -> None:
        """Generates braille for contents and displays the flattened regions."""

        if not self.use_braille():
            return

        context = self._build_generator_context()
        regions_list, focused_region = script.get_braille_generator().generate_contents(
            contents,
            context,
        )
        if not regions_list:
            return

        flattened_regions: list = []
        for regions in regions_list:
            flattened_regions.extend(regions)
        self.present_regions(flattened_regions, focused_region)

    def present_generated_braille(  # pylint: disable=too-many-arguments
        self,
        script: default.Script,
        obj: Atspi.Accessible,
        *,
        prior_obj: Atspi.Accessible | None = None,
        reason: PresentationReason | None = None,
        offset: int | None = None,
    ) -> None:
        """Generates braille for obj using the script's braille generator and displays it."""

        if not self.use_braille():
            return

        context = self._build_generator_context(
            reason,
            prior_obj=prior_obj,
            offset=offset,
        )
        generator = script.get_braille_generator()
        result, focused_region = generator.generate_braille(obj, context)
        if result:
            self.present_regions(list(result), focused_region)

    def _process_braille_output(
        self,
        regions: list[braille.Region],
        focused_region: braille.Region | None,
    ) -> tuple[list[braille.Region], braille.Region | None, bool]:
        """Lets extensions observe, replace, or consume outgoing braille."""

        handlers = extension_loader.get_loader().iter_braille_output_handlers()
        if not handlers:
            return regions, focused_region, False

        tokens = [
            "BRAILLE OUTPUT HOOK: regions:",
            regions,
            "focused region:",
            focused_region,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        output = BrailleOutput(tuple(regions), focused_region)
        for handler in handlers:
            result = self._call_braille_output_handler(handler, output)
            if result is None:
                continue
            if result.text is not None:
                tokens = [
                    "BRAILLE OUTPUT HOOK: Extension",
                    handler.module_name,
                    "replaced text:",
                    result.text,
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                regions = [braille.Region(result.text)]
                focused_region = regions[0] if result.text else None
                output = BrailleOutput(tuple(regions), focused_region)
            if result.consume:
                tokens = [
                    "BRAILLE OUTPUT HOOK: Extension",
                    handler.module_name,
                    "consumed output.",
                ]
                debug.print_tokens(debug.LEVEL_INFO, tokens, True)
                return regions, focused_region, True

        return regions, focused_region, False

    @staticmethod
    def _call_braille_output_handler(
        handler: Extension,
        output: BrailleOutput,
    ) -> BrailleOutputResult | None:
        """Calls a braille output handler and validates the result."""

        tokens = ["BRAILLE OUTPUT HOOK: Calling extension:", handler.module_name]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        try:
            result = handler.on_braille_output(output)
        except Exception as error:  # pylint: disable=broad-exception-caught
            msg = (
                f"BRAILLE PRESENTER: Extension {handler.module_name} "
                f"failed while handling braille output: {error}"
            )
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return None

        if result is None:
            return None
        if not isinstance(result, BrailleOutputResult):
            msg = (
                f"BRAILLE PRESENTER: Extension {handler.module_name} "
                f"returned unexpected braille output result: {result}"
            )
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return None
        if result.text is not None and not isinstance(result.text, str):
            msg = (
                f"BRAILLE PRESENTER: Extension {handler.module_name} "
                f"returned non-string braille text: {result.text}"
            )
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return None
        return result

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ENABLED,
        schema="braille",
        gtype="b",
        default=True,
        summary="Enable braille output",
        migration_key="enableBraille",
    )
    @dbus_service.getter
    def get_braille_is_enabled(self) -> bool:
        """Returns whether braille is enabled."""

        return self._get_setting(self.KEY_ENABLED, "b", True)

    @dbus_service.setter
    def set_braille_is_enabled(self, value: bool) -> bool:
        """Sets whether braille is enabled."""

        if value == self.get_braille_is_enabled():
            return True

        msg = f"BRAILLE PRESENTER: Setting enable braille to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(self._SCHEMA, self.KEY_ENABLED, value)
        braille.set_enable_braille(value)

        if value:
            braille.init(input_event_manager.get_manager().process_braille_event)
        else:
            braille.shutdown()

        return True

    def use_verbose_braille(self) -> bool:
        """Returns whether the braille verbosity level is set to verbose."""

        return self.get_verbosity_level() == "verbose"

    @gsettings_registry.get_registry().gsetting(
        key=KEY_VERBOSITY_LEVEL,
        schema="braille",
        genum="org.gnome.Orca.VerbosityLevel",
        default="verbose",
        summary="Braille verbosity level (brief, verbose)",
        migration_key="brailleVerbosityLevel",
    )
    @dbus_service.getter
    def get_verbosity_level(self) -> str:
        """Returns the current braille verbosity level for object presentation."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_VERBOSITY_LEVEL,
            "",
            genum="org.gnome.Orca.VerbosityLevel",
            default="verbose",
        )

    @dbus_service.setter
    def set_verbosity_level(self, value: str) -> bool:
        """Sets the braille verbosity level for object presentation."""

        try:
            level = VerbosityLevel[value.upper()]
        except KeyError:
            msg = f"BRAILLE PRESENTER: Invalid verbosity level: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        msg = f"BRAILLE PRESENTER: Setting verbosity level to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_VERBOSITY_LEVEL,
            level.string_name,
        )
        return True

    def use_full_rolenames(self) -> bool:
        """Returns whether full rolenames should be used."""

        return self.get_rolename_style() == "verbose"

    @gsettings_registry.get_registry().gsetting(
        key=KEY_ROLENAME_STYLE,
        schema="braille",
        genum="org.gnome.Orca.VerbosityLevel",
        default="verbose",
        summary="Braille rolename style (brief, verbose)",
        migration_key="brailleRolenameStyle",
    )
    @dbus_service.getter
    def get_rolename_style(self) -> str:
        """Returns the current rolename style for object presentation."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_ROLENAME_STYLE,
            "",
            genum="org.gnome.Orca.VerbosityLevel",
            default="verbose",
        )

    @dbus_service.setter
    def set_rolename_style(self, value: str) -> bool:
        """Sets the current rolename style for object presentation."""

        try:
            level = VerbosityLevel[value.upper()]
        except KeyError:
            msg = f"BRAILLE PRESENTER: Invalid rolename style: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        msg = f"BRAILLE PRESENTER: Setting rolename style to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_ROLENAME_STYLE,
            level.string_name,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PRESENT_MNEMONICS,
        schema="braille",
        gtype="b",
        default=True,
        summary="Present mnemonics on braille display",
        migration_key="displayObjectMnemonic",
    )
    @dbus_service.getter
    def get_present_mnemonics(self) -> bool:
        """Returns whether mnemonics are presented on the braille display."""

        return self._get_setting(self.KEY_PRESENT_MNEMONICS, "b", True)

    @dbus_service.setter
    def set_present_mnemonics(self, value: bool) -> bool:
        """Sets whether mnemonics are presented on the braille display."""

        msg = f"BRAILLE PRESENTER: Setting present mnemonics to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_PRESENT_MNEMONICS,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_DISPLAY_ANCESTORS,
        schema="braille",
        gtype="b",
        default=True,
        summary="Display ancestors of current object",
        migration_key="enableBrailleContext",
    )
    @dbus_service.getter
    def get_display_ancestors(self) -> bool:
        """Returns whether ancestors of the current object will be displayed."""

        return self._get_setting(self.KEY_DISPLAY_ANCESTORS, "b", True)

    @dbus_service.setter
    def set_display_ancestors(self, value: bool) -> bool:
        """Sets whether ancestors of the current object will be displayed."""

        msg = f"BRAILLE PRESENTER: Setting enable braille context to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_DISPLAY_ANCESTORS,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_BRAILLE_PROGRESS_BAR_UPDATES,
        schema="braille",
        gtype="b",
        default=False,
        summary="Show progress bar updates in braille",
        migration_key="brailleProgressBarUpdates",
    )
    @dbus_service.getter
    def get_braille_progress_bar_updates(self) -> bool:
        """Returns whether braille progress bar updates are enabled."""

        return self._get_setting(self.KEY_BRAILLE_PROGRESS_BAR_UPDATES, "b", False)

    @dbus_service.setter
    def set_braille_progress_bar_updates(self, value: bool) -> bool:
        """Sets whether braille progress bar updates are enabled."""

        msg = f"BRAILLE PRESENTER: Setting braille progress bar updates to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_BRAILLE_PROGRESS_BAR_UPDATES,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PROGRESS_BAR_BRAILLE_INTERVAL,
        schema="braille",
        gtype="i",
        default=10,
        summary="Progress bar braille update interval in seconds",
        migration_key="progressBarBrailleInterval",
    )
    @dbus_service.getter
    def get_progress_bar_braille_interval(self) -> UInt32:
        """Returns the braille progress bar update interval in seconds."""

        return self._get_setting(self.KEY_PROGRESS_BAR_BRAILLE_INTERVAL, "i", 10)

    @dbus_service.setter
    def set_progress_bar_braille_interval(self, value: UInt32) -> bool:
        """Sets the braille progress bar update interval in seconds."""

        msg = f"BRAILLE PRESENTER: Setting progress bar braille interval to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_PROGRESS_BAR_BRAILLE_INTERVAL,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_PROGRESS_BAR_BRAILLE_VERBOSITY,
        schema="braille",
        genum="org.gnome.Orca.ProgressBarVerbosity",
        default="application",
        summary="Progress bar braille verbosity (all, application, window)",
        migration_key="progressBarBrailleVerbosity",
    )
    @dbus_service.getter
    def get_progress_bar_braille_verbosity(self) -> UInt32:
        """Returns the braille progress bar verbosity level."""

        nick = gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_PROGRESS_BAR_BRAILLE_VERBOSITY,
            "",
            genum="org.gnome.Orca.ProgressBarVerbosity",
            default="application",
        )
        return ProgressBarVerbosity[nick.upper()].value

    @dbus_service.setter
    def set_progress_bar_braille_verbosity(self, value: UInt32) -> bool:
        """Sets the braille progress bar verbosity level."""

        msg = f"BRAILLE PRESENTER: Setting progress bar braille verbosity to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        level = ProgressBarVerbosity(value)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_PROGRESS_BAR_BRAILLE_VERBOSITY,
            level.name.lower(),
        )
        return True

    def should_present_progress_bar_update(
        self,
        obj: Atspi.Accessible,
        percent: int | None,
        is_same_app: bool,
        is_same_window: bool,
    ) -> bool:
        """Returns True if the progress bar update should be brailled."""

        if not self.get_braille_progress_bar_updates():
            return False

        last_time, last_value = self._progress_bar_cache.get(id(obj), (0.0, None))
        if percent == last_value:
            return False

        if percent != 100:
            interval = int(time.time() - last_time)
            if interval < self.get_progress_bar_braille_interval():
                return False

        verbosity = self.get_progress_bar_braille_verbosity()
        if verbosity == ProgressBarVerbosity.ALL.value:
            present = True
        elif verbosity == ProgressBarVerbosity.APPLICATION.value:
            present = is_same_app
        elif verbosity == ProgressBarVerbosity.WINDOW.value:
            present = is_same_window
        else:
            present = True

        if present:
            self._progress_bar_cache[id(obj)] = (time.time(), percent)

        return present

    @gsettings_registry.get_registry().gsetting(
        key=KEY_CONTRACTED_BRAILLE,
        schema="braille",
        gtype="b",
        default=False,
        summary="Enable contracted braille",
        migration_key="enableContractedBraille",
    )
    @dbus_service.getter
    def get_contracted_braille_is_enabled(self) -> bool:
        """Returns whether contracted braille is enabled."""

        return self._get_setting(self.KEY_CONTRACTED_BRAILLE, "b", False)

    @dbus_service.setter
    def set_contracted_braille_is_enabled(self, value: bool) -> bool:
        """Sets whether contracted braille is enabled."""

        msg = f"BRAILLE PRESENTER: Setting enable contracted braille to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_CONTRACTED_BRAILLE,
            value,
        )
        braille.set_enable_contracted_braille(value)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_COMPUTER_BRAILLE_AT_CURSOR,
        schema="braille",
        gtype="b",
        default=True,
        summary="Use computer braille at cursor position",
        migration_key="enableComputerBrailleAtCursor",
    )
    @dbus_service.getter
    def get_computer_braille_at_cursor_is_enabled(self) -> bool:
        """Returns whether computer braille is used at the cursor position."""

        return self._get_setting(self.KEY_COMPUTER_BRAILLE_AT_CURSOR, "b", True)

    @dbus_service.setter
    def set_computer_braille_at_cursor_is_enabled(self, value: bool) -> bool:
        """Sets whether computer braille is used at the cursor position."""

        msg = f"BRAILLE PRESENTER: Setting enable computer braille at cursor to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_COMPUTER_BRAILLE_AT_CURSOR,
            value,
        )
        braille.set_enable_computer_braille_at_cursor(value)
        return True

    def get_contraction_table_path(self) -> str:
        """Returns the current braille contraction table file path."""

        value = gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_CONTRACTION_TABLE,
            "s",
            default="",
        )
        return value or braille.get_default_contraction_table()

    @gsettings_registry.get_registry().gsetting(
        key=KEY_CONTRACTION_TABLE,
        schema="braille",
        gtype="s",
        default="",
        summary="Braille contraction table name",
        migration_key="brailleContractionTable",
    )
    @dbus_service.getter
    def get_contraction_table(self) -> str:
        """Returns the current braille contraction table name."""

        full_path = gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_CONTRACTION_TABLE,
            "s",
            default="",
        )
        if not full_path:
            return ""
        return os.path.splitext(os.path.basename(full_path))[0]

    @dbus_service.getter
    def get_available_contraction_tables(self) -> list[str]:
        """Returns a list of available contraction table names."""

        table_files = self._get_table_files()
        return [os.path.splitext(filename)[0] for filename in table_files]

    def get_contraction_tables_dict(self) -> dict[str, str]:
        """Returns a dictionary mapping display names to table file paths."""

        names = self.get_table_names()
        tables = {}
        for fname in self._get_table_files():
            display_name = names.get(fname, os.path.splitext(fname)[0])
            tables[display_name] = os.path.join(tablesdir, fname)
        return tables

    @dbus_service.setter
    def set_contraction_table(self, value: str) -> bool:
        """Sets the current braille contraction table."""

        table_files = self._get_table_files()
        base_name = os.path.splitext(value)[0]
        filename = None
        for table_file in table_files:
            if table_file.startswith(base_name + "."):
                filename = table_file
                break

        if not filename:
            msg = f"BRAILLE PRESENTER: Invalid contraction table: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        full_path = os.path.join(tablesdir, filename)
        return self.set_contraction_table_from_path(full_path)

    @staticmethod
    def _get_table_files() -> list[str]:
        """Returns a list of braille table filenames in the tables directory."""

        try:
            return [fname for fname in os.listdir(tablesdir) if fname[-4:] in (".utb", ".ctb")]
        except OSError:
            return []

    def set_contraction_table_from_path(self, file_path: str) -> bool:
        """Sets the current braille contraction table from a file path."""

        msg = f"BRAILLE PRESENTER: Setting contraction table to {file_path}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_CONTRACTION_TABLE,
            file_path,
        )
        braille.set_contraction_table(file_path)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_END_OF_LINE_INDICATOR,
        schema="braille",
        gtype="b",
        default=True,
        summary="Show end-of-line indicator",
        migration_key="enableBrailleEOL",
    )
    @dbus_service.getter
    def get_end_of_line_indicator_is_enabled(self) -> bool:
        """Returns whether the end-of-line indicator is enabled."""

        return self._get_setting(self.KEY_END_OF_LINE_INDICATOR, "b", True)

    @dbus_service.setter
    def set_end_of_line_indicator_is_enabled(self, value: bool) -> bool:
        """Sets whether the end-of-line indicator is enabled."""

        msg = f"BRAILLE PRESENTER: Setting enable-eol to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_END_OF_LINE_INDICATOR,
            value,
        )
        braille.set_enable_eol(value)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_WORD_WRAP,
        schema="braille",
        gtype="b",
        default=False,
        summary="Enable braille word wrap",
        migration_key="enableBrailleWordWrap",
    )
    @dbus_service.getter
    def get_word_wrap_is_enabled(self) -> bool:
        """Returns whether braille word wrap is enabled."""

        return self._get_setting(self.KEY_WORD_WRAP, "b", False)

    @dbus_service.setter
    def set_word_wrap_is_enabled(self, value: bool) -> bool:
        """Sets whether braille word wrap is enabled."""

        msg = f"BRAILLE PRESENTER: Setting enable word wrap to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(self._SCHEMA, self.KEY_WORD_WRAP, value)
        braille.set_enable_word_wrap(value)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FLASH_MESSAGES,
        schema="braille",
        gtype="b",
        default=True,
        summary="Enable braille flash messages",
        migration_key="enableFlashMessages",
    )
    @dbus_service.getter
    def get_flash_messages_are_enabled(self) -> bool:
        """Returns whether 'flash' messages (i.e. announcements) are enabled."""

        return self._get_setting(self.KEY_FLASH_MESSAGES, "b", True)

    @dbus_service.setter
    def set_flash_messages_are_enabled(self, value: bool) -> bool:
        """Sets whether 'flash' messages (i.e. announcements) are enabled."""

        msg = f"BRAILLE PRESENTER: Setting enable flash messages to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA, self.KEY_FLASH_MESSAGES, value
        )
        return True

    def get_flashtime_from_settings(self) -> int:
        """Returns flash message duration in milliseconds based on user settings."""

        if self.get_flash_messages_are_persistent():
            return -1
        return self.get_flash_message_duration()

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FLASH_MESSAGE_DURATION,
        schema="braille",
        gtype="i",
        default=5000,
        summary="Flash message duration in milliseconds",
        migration_key="brailleFlashTime",
    )
    @dbus_service.getter
    def get_flash_message_duration(self) -> UInt32:
        """Returns flash message duration in milliseconds."""

        return self._get_setting(self.KEY_FLASH_MESSAGE_DURATION, "i", 5000)

    @dbus_service.setter
    def set_flash_message_duration(self, value: UInt32) -> bool:
        """Sets flash message duration in milliseconds."""

        msg = f"BRAILLE PRESENTER: Setting braille flash time to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_FLASH_MESSAGE_DURATION,
            value,
        )
        return True

    def set_selector_indicator_from_int(self, value: int) -> bool:
        """Sets the braille selector indicator from an int value."""

        msg = f"BRAILLE PRESENTER: Setting selector indicator to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        indicator = BrailleIndicator(value)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_SELECTOR_INDICATOR,
            indicator.string_name,
        )
        braille.set_selector_indicator(value)
        return True

    def set_link_indicator_from_int(self, value: int) -> bool:
        """Sets the braille link indicator from an int value."""

        msg = f"BRAILLE PRESENTER: Setting link indicator to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        indicator = BrailleIndicator(value)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_LINK_INDICATOR,
            indicator.string_name,
        )
        braille.set_link_indicator(value)
        return True

    def set_text_attributes_indicator_from_int(self, value: int) -> bool:
        """Sets the braille text attributes indicator from an int value."""

        msg = f"BRAILLE PRESENTER: Setting text attributes indicator to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        indicator = BrailleIndicator(value)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_TEXT_ATTRIBUTES_INDICATOR,
            indicator.string_name,
        )
        braille.set_text_attributes_indicator(value)
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FLASH_MESSAGES_PERSISTENT,
        schema="braille",
        gtype="b",
        default=False,
        summary="Make flash messages persistent",
        migration_key="flashIsPersistent",
    )
    @dbus_service.getter
    def get_flash_messages_are_persistent(self) -> bool:
        """Returns whether 'flash' messages are persistent (as opposed to temporary)."""

        return self._get_setting(self.KEY_FLASH_MESSAGES_PERSISTENT, "b", False)

    @dbus_service.setter
    def set_flash_messages_are_persistent(self, value: bool) -> bool:
        """Sets whether 'flash' messages are persistent (as opposed to temporary)."""

        msg = f"BRAILLE PRESENTER: Setting flash messages are persistent to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_FLASH_MESSAGES_PERSISTENT,
            value,
        )
        return True

    @gsettings_registry.get_registry().gsetting(
        key=KEY_FLASH_MESSAGES_DETAILED,
        schema="braille",
        gtype="b",
        default=True,
        summary="Use detailed flash messages",
        migration_key="flashIsDetailed",
    )
    @dbus_service.getter
    def get_flash_messages_are_detailed(self) -> bool:
        """Returns whether 'flash' messages are detailed (as opposed to brief)."""

        return self._get_setting(self.KEY_FLASH_MESSAGES_DETAILED, "b", True)

    @dbus_service.setter
    def set_flash_messages_are_detailed(self, value: bool) -> bool:
        """Sets whether 'flash' messages are detailed (as opposed to brief)."""

        msg = f"BRAILLE PRESENTER: Setting flash messages are detailed to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_FLASH_MESSAGES_DETAILED,
            value,
        )
        return True

    def _get_selector_indicator_as_int(self) -> int:
        """Returns the braille selector indicator as an int."""

        nick = gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_SELECTOR_INDICATOR,
            "",
            genum="org.gnome.Orca.BrailleIndicator",
            default="dots78",
        )
        value = BrailleIndicator[nick.upper()].value
        msg = f"BRAILLE PRESENTER: Getting selector indicator: {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return value

    @gsettings_registry.get_registry().gsetting(
        key=KEY_SELECTOR_INDICATOR,
        schema="braille",
        genum="org.gnome.Orca.BrailleIndicator",
        default="dots78",
        summary="Braille selector indicator style (none, dot7, dot8, dots78)",
        migration_key="brailleSelectorIndicator",
    )
    @dbus_service.getter
    def get_selector_indicator(self) -> str:
        """Returns the braille selector indicator style."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_SELECTOR_INDICATOR,
            "",
            genum="org.gnome.Orca.BrailleIndicator",
            default="dots78",
        )

    @dbus_service.setter
    def set_selector_indicator(self, value: str) -> bool:
        """Sets the braille selector indicator style."""

        try:
            indicator = BrailleIndicator[value.upper()]
        except KeyError:
            msg = f"BRAILLE PRESENTER: Invalid selector indicator: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        msg = f"BRAILLE PRESENTER: Setting selector indicator to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_SELECTOR_INDICATOR,
            indicator.string_name,
        )
        braille.set_selector_indicator(indicator.value)
        return True

    def _get_link_indicator_as_int(self) -> int:
        """Returns the braille link indicator as an int."""

        nick = gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_LINK_INDICATOR,
            "",
            genum="org.gnome.Orca.BrailleIndicator",
            default="dots78",
        )
        value = BrailleIndicator[nick.upper()].value
        msg = f"BRAILLE PRESENTER: Getting link indicator: {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return value

    @gsettings_registry.get_registry().gsetting(
        key=KEY_LINK_INDICATOR,
        schema="braille",
        genum="org.gnome.Orca.BrailleIndicator",
        default="dots78",
        summary="Braille link indicator style (none, dot7, dot8, dots78)",
        migration_key="brailleLinkIndicator",
    )
    @dbus_service.getter
    def get_link_indicator(self) -> str:
        """Returns the braille link indicator style."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_LINK_INDICATOR,
            "",
            genum="org.gnome.Orca.BrailleIndicator",
            default="dots78",
        )

    @dbus_service.setter
    def set_link_indicator(self, value: str) -> bool:
        """Sets the braille link indicator style."""

        try:
            indicator = BrailleIndicator[value.upper()]
        except KeyError:
            msg = f"BRAILLE PRESENTER: Invalid link indicator: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        msg = f"BRAILLE PRESENTER: Setting link indicator to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_LINK_INDICATOR,
            indicator.string_name,
        )
        braille.set_link_indicator(indicator.value)
        return True

    def _get_text_attributes_indicator_as_int(self) -> int:
        """Returns the braille text attributes indicator as an int."""

        nick = gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_TEXT_ATTRIBUTES_INDICATOR,
            "",
            genum="org.gnome.Orca.BrailleIndicator",
            default="none",
        )
        value = BrailleIndicator[nick.upper()].value
        msg = f"BRAILLE PRESENTER: Getting text attributes indicator: {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return value

    @gsettings_registry.get_registry().gsetting(
        key=KEY_TEXT_ATTRIBUTES_INDICATOR,
        schema="braille",
        genum="org.gnome.Orca.BrailleIndicator",
        default="none",
        summary="Braille text attributes indicator style (none, dot7, dot8, dots78)",
        migration_key="textAttributesBrailleIndicator",
    )
    @dbus_service.getter
    def get_text_attributes_indicator(self) -> str:
        """Returns the braille text attributes indicator style."""

        return gsettings_registry.get_registry().layered_lookup(
            self._SCHEMA,
            self.KEY_TEXT_ATTRIBUTES_INDICATOR,
            "",
            genum="org.gnome.Orca.BrailleIndicator",
            default="none",
        )

    @dbus_service.setter
    def set_text_attributes_indicator(self, value: str) -> bool:
        """Sets the braille text attributes indicator style."""

        try:
            indicator = BrailleIndicator[value.upper()]
        except KeyError:
            msg = f"BRAILLE PRESENTER: Invalid text attributes indicator: {value}"
            debug.print_message(debug.LEVEL_WARNING, msg, True)
            return False

        msg = f"BRAILLE PRESENTER: Setting text attributes indicator to {value}."
        debug.print_message(debug.LEVEL_INFO, msg, True)
        gsettings_registry.get_registry().set_runtime_value(
            self._SCHEMA,
            self.KEY_TEXT_ATTRIBUTES_INDICATOR,
            indicator.string_name,
        )
        braille.set_text_attributes_indicator(indicator.value)
        return True

    def is_flash_active(self) -> bool:
        """Returns True if a flash message is currently being displayed."""

        return braille.is_flash_active()

    def kill_flash(self, restore_saved: bool = True) -> None:
        """Kills any flashed message currently being displayed."""

        braille.kill_flash(restore_saved)

    def present_message(
        self,
        message: str,
        restore_previous: bool = True,
        flash_time: int | None = None,
    ) -> None:
        """Displays a single line message in braille."""

        if not self.use_braille():
            return

        if flash_time is None:
            flash_time = self.get_flashtime_from_settings()

        if not restore_previous and flash_time:
            braille.kill_flash(restore_saved=False)

        braille.display_message(message, flash_time=flash_time)

    def pan_left(self) -> bool:
        """Pans the braille display left, returning True if the display moved."""

        if not self.use_braille():
            return False

        moved = braille.pan_left()
        if moved:
            braille.refresh(pan_to_cursor=False, stop_flash=False)
        return moved

    def pan_right(self) -> bool:
        """Pans the braille display right, returning True if the display moved."""

        if not self.use_braille():
            return False

        moved = braille.pan_right()
        if moved:
            braille.refresh(pan_to_cursor=False, stop_flash=False)
        return moved

    def refresh_braille(self, pan_to_cursor: bool = True) -> None:
        """Refreshes the braille display without rebuilding the line."""

        if not self.use_braille():
            return

        braille.refresh(pan_to_cursor=pan_to_cursor)

    def pan_to_beginning(self) -> None:
        """Pans the braille display all the way to the beginning of the line."""

        if not self.use_braille():
            return

        while braille.pan_left():
            pass
        braille.refresh(pan_to_cursor=False, stop_flash=True)

    def pan_to_end(self) -> None:
        """Pans the braille display all the way to the end of the line."""

        if not self.use_braille():
            return

        while braille.pan_right():
            pass
        braille.refresh(pan_to_cursor=False, stop_flash=True)

    def check_braille_setting(self) -> None:
        """Checks the braille setting and disables braille if necessary."""

        braille.check_braille_setting()

    def disable_braille(self) -> None:
        """Idles or shuts down braille output if enabled."""

        braille.disable_braille()

    def set_brlapi_priority(self, high: bool = False) -> None:
        """Sets the BrlAPI priority level.

        Args:
            high: If True, use high priority (for flat review). Otherwise use default.
        """

        if high:
            braille.set_brlapi_priority(braille.BRLAPI_PRIORITY_HIGH)
        else:
            braille.set_brlapi_priority()

    def update_monitor(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        cursor_cell: int,
        visible: str,
        visible_mask: str | None,
        display_size: int,
        full: str,
        full_mask: str | None,
    ) -> None:
        """Updates the braille monitor display, creating it on demand if enabled."""

        self._output_recorder.record(
            kind="braille",
            cursor_cell=cursor_cell,
            full=full,
            visible=visible,
            mask=full_mask,
        )
        if not self.get_monitor_is_enabled():
            return

        if braille.has_braille_device():
            cell_count = display_size
        else:
            cell_count = self.get_monitor_cell_count() or display_size
        if self._monitor is not None and len(self._monitor.cells) != cell_count:
            self.destroy_monitor()
        if self._monitor is None:
            self._monitor = braille_monitor.BrailleMonitor(
                cell_count,
                on_close=lambda: self.set_monitor_is_enabled(False),
                foreground=self.get_monitor_foreground(),
                background=self.get_monitor_background(),
            )
            self._monitor.show_all()  # pylint: disable=no-member

        if self.get_monitor_show_dots():
            visible = self._to_unicode_braille(visible)

        self._monitor.write_text(cursor_cell, visible, visible_mask)

    def _to_unicode_braille(self, text: str) -> str:
        """Convert text to Unicode braille dot pattern characters.

        Uses louis.charToDots() to map each character to its braille dot pattern,
        then converts to Unicode braille characters (U+2800 block).
        """

        try:
            import louis  # pylint: disable=import-outside-toplevel

            table = (
                self.get_contraction_table_path()
                if self.get_contracted_braille_is_enabled()
                else ""
            )
            if not table:
                table = "en-us-comp8.ctb"
            dots_str = louis.charToDots([table], text)
            return "".join(chr(0x2800 | (ord(c) & 0xFF)) for c in dots_str)
        except Exception:  # pylint: disable=broad-except
            return text

    def destroy_monitor(self) -> None:
        """Destroys the braille monitor widget if it exists."""

        if self._monitor is not None:
            self._monitor.destroy()
            self._monitor = None

    def init_braille(self) -> None:
        """Initializes braille if enabled."""

        braille.set_monitor_callback(self.update_monitor)
        if not self.get_braille_is_enabled():
            return

        braille.init(input_event_manager.get_manager().process_braille_event)
        self._sync_state_to_braille()

    def _sync_state_to_braille(self) -> None:
        """Pushes persisted dconf settings into the braille module's runtime state."""

        braille.set_enable_contracted_braille(self.get_contracted_braille_is_enabled())
        braille.set_contraction_table(self.get_contraction_table_path())
        braille.set_enable_computer_braille_at_cursor(
            self.get_computer_braille_at_cursor_is_enabled(),
        )
        braille.set_enable_eol(self.get_end_of_line_indicator_is_enabled())
        braille.set_enable_word_wrap(self.get_word_wrap_is_enabled())
        braille.set_selector_indicator(self._get_selector_indicator_as_int())
        braille.set_link_indicator(self._get_link_indicator_as_int())
        braille.set_text_attributes_indicator(self._get_text_attributes_indicator_as_int())

    def shutdown_braille(self) -> None:
        """Shuts down braille."""

        braille.shutdown()


_presenter: BraillePresenter = BraillePresenter()


def get_presenter() -> BraillePresenter:
    """Returns the Braille Presenter"""

    return _presenter



# Orca
#
# Copyright 2023-2026 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Utilities for accessible actions."""

from __future__ import annotations

import gi

gi.require_version("Atspi", "2.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Atspi, GLib, Gtk

from . import debug, keynames
from .ax_action import AXAction
from .ax_object import AXObject


class AXUtilitiesAction:
    """Utilities for accessible actions."""

    @staticmethod
    def get_action_names(obj: Atspi.Accessible) -> list[str]:
        """Returns the list of actions supported on obj."""

        results = []
        n_actions = AXAction.get_n_actions(obj)
        for i in range(n_actions):
            name = AXAction.get_action_name(obj, i, n_actions)
            if name:
                results.append(name)
        return results

    @staticmethod
    def get_action_index(obj: Atspi.Accessible, action_name: str) -> int:
        """Returns the index of the named action or -1 if unsupported."""

        action_name = AXAction.normalize_action_name(action_name)
        n_actions = AXAction.get_n_actions(obj)
        for i in range(n_actions):
            if action_name == AXAction.get_action_name(obj, i, n_actions):
                return i

        return -1

    @staticmethod
    def has_action(obj: Atspi.Accessible, action_name: str) -> bool:
        """Returns true if the named action is supported on obj."""

        return AXUtilitiesAction.get_action_index(obj, action_name) >= 0

    @staticmethod
    def do_named_action(obj: Atspi.Accessible, action_name: str) -> bool:
        """Invokes the named action on obj. The return value, if true, may be
        meaningless because most implementors return true without knowing if
        the action was successfully performed."""

        index = AXUtilitiesAction.get_action_index(obj, action_name)
        if index == -1:
            tokens = ["INFO:", action_name, "not an available action for", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return False

        return AXAction.do_action(obj, index)

    @staticmethod
    def has_matching_shortcut(obj: Atspi.Accessible, key: str) -> bool:
        """Returns True if any action shortcut on obj matches key."""

        if not key:
            return False

        n_actions = AXAction.get_n_actions(obj)
        for i in range(n_actions):
            shortcuts = AXAction.get_action_key_binding(obj, i, n_actions).split(";")
            if any(s.endswith(key.upper()) for s in shortcuts):
                return True
        return False

    @staticmethod
    def _find_first_action_with_keybinding(obj: Atspi.Accessible) -> int:
        """Returns the index of the first action with a keybinding on obj."""

        n_actions = AXAction.get_n_actions(obj)
        for i in range(n_actions):
            if AXAction.get_action_key_binding(obj, i, n_actions):
                return i
        return -1

    @staticmethod
    def _get_label_for_key_sequence(sequence: str) -> str:
        """Returns the human consumable label for the key sequence."""

        if not sequence:
            return ""

        # We get all sorts of variations in the keybinding string. Try to normalize it.
        if len(sequence) > 1 and not sequence.startswith("<") and "," not in sequence:
            tokens = sequence.split("+")
            sequence = "".join(f"<{part}>" for part in tokens[:-1]) + tokens[-1]

        # We use Gtk for conversion to handle things like <Primary>.
        try:
            key, mods = Gtk.accelerator_parse(sequence)
            result = Gtk.accelerator_get_label(key, mods)
        except GLib.GError as error:
            msg = f"AXUtilitiesAction: Exception in _get_label_for_key_sequence: {error}"
            debug.print_message(debug.LEVEL_INFO, msg, True)
            sequence = sequence.replace("<", "").replace(">", " ").strip()
        else:
            if result and not result.endswith("+"):
                sequence = result

        return keynames.localize_key_sequence(sequence)

    @staticmethod
    def get_accelerator(obj: Atspi.Accessible) -> str:
        """Returns the accelerator/shortcut associated with obj."""

        attrs = AXObject.get_attributes_dict(obj)
        # The ARIA spec suggests a given shortcut's components should be separated by a "+".
        # Multiple shortcuts are apparently allowed and separated by a space.
        shortcuts = attrs.get("keyshortcuts", "").split(" ")
        if shortcuts and shortcuts[0]:
            result = " ".join(
                map(AXUtilitiesAction._get_label_for_key_sequence, shortcuts),
            ).strip()
            # Accelerators are typically modified and thus more than one character.
            if len(result) > 1:
                return result

        index = AXUtilitiesAction._find_first_action_with_keybinding(obj)
        if index == -1:
            return ""

        # This should be a string separated by semicolons and in the form:
        #     <mnemonic>;<full sequence>;<accelerator/shortcut> (optional)
        # In practice we get all sorts of variations.

        # If there's a third item, it's probably the accelerator.
        strings = AXAction.get_action_key_binding(obj, index).split(";")
        if len(strings) == 3:
            return AXUtilitiesAction._get_label_for_key_sequence(strings[2])

        # If the last thing has Ctrl in it, it's probably the accelerator.
        result = AXUtilitiesAction._get_label_for_key_sequence(strings[-1])
        if "Ctrl" in result:
            return result

        return ""

    @staticmethod
    def get_mnemonic(obj: Atspi.Accessible) -> str:
        """Returns the mnemonic associated with obj."""

        attrs = AXObject.get_attributes_dict(obj)
        # The ARIA spec suggests a given shortcut's components should be separated by a "+".
        # Multiple shortcuts are apparently allowed and separated by a space.
        shortcuts = attrs.get("keyshortcuts", "").split(" ")
        if shortcuts and shortcuts[0]:
            result = " ".join(
                map(AXUtilitiesAction._get_label_for_key_sequence, shortcuts),
            ).strip()
            # If it's not a single letter it's probably not the mnemonic.
            if len(result) == 1:
                return result

        index = AXUtilitiesAction._find_first_action_with_keybinding(obj)
        if index == -1:
            return ""

        # This should be a string separated by semicolons and in the form:
        #     <mnemonic>;<full sequence>;<accelerator/shortcut> (optional)
        # In practice we get all sorts of variations.

        strings = AXAction.get_action_key_binding(obj, index).split(";")
        result = AXUtilitiesAction._get_label_for_key_sequence(strings[0])
        # If Ctrl is in the result, it's probably the accelerator rather than the mnemonic.
        if "Ctrl" in result or "Control" in result:
            return ""

        # Don't treat space as a mnemonic.
        if result.lower() in [" ", "space", "<space>"]:
            return ""

        return result



# Orca
#
# Copyright 2023 Igalia, S.L.
# Author: Joanmarie Diggs <jdiggs@igalia.com>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

# pylint: disable=too-many-positional-arguments

"""Wrapper for the Atspi.Collection interface."""

from __future__ import annotations

import time

import gi

gi.require_version("Atspi", "2.0")
from gi.repository import Atspi, GLib

from . import debug
from .ax_object import AXObject


class AXCollection:
    """Wrapper for the Atspi.Collection interface."""

    # Too many arguments and too many local variables.
    # This function wraps Atspi.MatchRule.new which has all the arguments.
    # pylint: disable=R0913,R0914
    @staticmethod
    def create_match_rule(
        states: list[str] | None = None,
        state_match_type: Atspi.CollectionMatchType = Atspi.CollectionMatchType.ALL,
        attributes: list[str] | None = None,
        attribute_match_type: Atspi.CollectionMatchType = Atspi.CollectionMatchType.ALL,
        roles: list[str] | None = None,
        role_match_type: Atspi.CollectionMatchType = Atspi.CollectionMatchType.ALL,
        interfaces: list[str] | None = None,
        interface_match_type: Atspi.CollectionMatchType = Atspi.CollectionMatchType.ALL,
        invert: bool = False,
    ) -> Atspi.MatchRule | None:
        """Creates a match rule based on the supplied criteria."""

        if states is None:
            states = []
        if attributes is None:
            attributes = []
        if roles is None:
            roles = []
        if interfaces is None:
            interfaces = []

        state_set = Atspi.StateSet()
        if states:
            for state in states:
                state_set.add(state)

        attributes_dict: dict[str, str] = {}
        if attributes:
            for attr in attributes:
                key, value = attr.split(":", 1)
                value = value.replace(":", r"\:")
                if key in attributes_dict:
                    attributes_dict[key] = attributes_dict[key] + ":" + value
                else:
                    attributes_dict[key] = value

        try:
            rule = Atspi.MatchRule.new(
                state_set,
                state_match_type,
                attributes_dict,
                attribute_match_type,
                roles,
                role_match_type,
                interfaces,
                interface_match_type,
                invert,
            )
        except GLib.GError as error:
            tokens = ["AXCollection: Exception in create_match_rule:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        return rule

    # pylint: enable=R0913,R0914

    @staticmethod
    def get_all_matches(
        obj: Atspi.Accessible,
        rule: Atspi.MatchRule,
        order: Atspi.CollectionSortOrder = Atspi.CollectionSortOrder.CANONICAL,
    ) -> list[Atspi.Accessible]:
        """Returns a list of objects matching the specified rule."""

        if not AXObject.supports_collection(obj):
            tokens = ["AXCollection:", obj, "does not implement this interface."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return []

        if rule is None:
            return []

        start = time.time()
        try:
            # 0 means no limit on the number of results
            # The final argument, traverse, is not supported but is expected.
            matches = Atspi.Collection.get_matches(obj, rule, order, 0, True)
        except GLib.GError as error:
            tokens = ["AXCollection: Exception in get_all_matches:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return []

        msg = f"AXCollection: {len(matches)} match(es) found in {time.time() - start:.4f}s"
        debug.print_message(debug.LEVEL_INFO, msg, True)
        return matches

    @staticmethod
    def get_first_match(
        obj: Atspi.Accessible,
        rule: Atspi.MatchRule,
        order: Atspi.CollectionSortOrder = Atspi.CollectionSortOrder.CANONICAL,
    ) -> Atspi.Accessible | None:
        """Returns the first object matching the specified rule."""

        if not AXObject.supports_collection(obj):
            tokens = ["AXCollection:", obj, "does not implement this interface."]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        if rule is None:
            return None

        start = time.time()
        try:
            # 1 means limit the number of results to 1
            # The final argument, traverse, is not supported but is expected.
            matches = Atspi.Collection.get_matches(obj, rule, order, 1, True)
        except GLib.GError as error:
            tokens = ["AXCollection: Exception in get_first_match:", error]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return None

        match = None
        if matches:
            match = matches[0]

        tokens = ["AXCollection: found", match, f"in {time.time() - start:.4f}s"]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return match



# Orca
#
# Copyright 2005-2008 Sun Microsystems Inc.
# Copyright 2016-2023 Igalia, S.L.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., Franklin Street, Fifth Floor,
# Boston MA  02110-1301 USA.

"""Module for commands related to the current accessible object."""

from __future__ import annotations

import time
from typing import TYPE_CHECKING

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, Gtk  # pylint: disable=no-name-in-module

from . import (
    dbus_service,
    debug,
    flat_review_presenter,
    focus_manager,
    guilabels,
    input_event,
    math_navigator,
    math_presenter,
    messages,
    presentation_manager,
    speech_presenter,
    spellcheck_presenter,
    text_attribute_manager,
    where_am_i_presenter_command_definitions,
)
from .ax_component import AXComponent
from .ax_object import AXObject
from .ax_text import AXText, AXTextAttribute
from .ax_utilities import AXUtilities
from .extension import Extension
from .generator import PresentationReason

if TYPE_CHECKING:
    gi.require_version("Atspi", "2.0")
    from gi.repository import Atspi

    from .command import Command
    from .scripts import default


class WhereAmIPresenter(Extension):
    """Module for commands related to the current accessible object."""

    GROUP_LABEL = guilabels.KB_GROUP_WHERE_AM_I

    def __init__(self) -> None:
        super().__init__()
        self._char_attributes_gui: CharacterAttributesGUI | None = None

    def _get_commands(self) -> list[Command]:
        return where_am_i_presenter_command_definitions.get_commands(self)

    def _get_current_character_info(self, script: default.Script) -> tuple[dict[str, str], str]:
        """Returns the text attributes and string for the current character."""

        reviewer = flat_review_presenter.get_presenter()
        if reviewer.is_active():
            context = reviewer.get_or_create_context(script)
            obj = context.get_current_object()
            offset = context.get_current_text_offset()
        else:
            obj = focus_manager.get_manager().get_locus_of_focus()
            offset = None

        attrs = AXText.get_text_attributes_at_offset(obj, offset)[0]
        char = AXText.get_character_at_offset(obj, offset)[0]
        return attrs, char

    def _localize_text_attribute(self, key: str, value: str | None) -> str:
        """Returns a localized description of the text attribute for Orca+F readout."""

        if value is None:
            return ""

        if key == "weight" and (value == "bold" or int(value) > 400):
            return messages.BOLD

        if key.endswith("spelling") or value == "spelling":
            return messages.MISSPELLED

        ax_text_attribute = AXTextAttribute.from_string(key)
        if ax_text_attribute is None:
            return ""

        localized_key = ax_text_attribute.get_localized_name()
        localized_value = ax_text_attribute.get_localized_value(value)
        return f"{localized_key}: {localized_value}"

    def _get_all_available_text_attributes(self, attrs: dict[str, str]) -> list[str]:
        """Returns localized descriptions of all available text attributes."""

        result = []
        seen = set()
        for key, value in attrs.items():
            ax_text_attribute = AXTextAttribute.from_string(key)
            sort_key = ax_text_attribute.get_localized_name() if ax_text_attribute else key
            canonical_key = ax_text_attribute.get_attribute_name() if ax_text_attribute else key

            if canonical_key in seen or value is None:
                continue

            if ax_text_attribute is not None:
                localized_key = ax_text_attribute.get_localized_name()
                localized_value = ax_text_attribute.get_localized_value(value)
                description = f"{localized_key}: {localized_value}"
            else:
                description = f"{key}: {value}"

            result.append((sort_key.casefold(), description))
            seen.add(canonical_key)

        return [description for _sort_key, description in sorted(result)]

    def _get_character_formatting_text(self, attrs: dict[str, str]) -> str:
        """Returns text for the character-formatting window."""

        return "\n".join(self._get_all_available_text_attributes(attrs))

    @dbus_service.command
    def present_character_attributes(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the font and formatting details for the current character."""

        tokens = [
            "WHERE AM I PRESENTER: present_character_attributes. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        attrs, _char = self._get_current_character_info(script)
        attr_list = text_attribute_manager.get_manager().get_resolved_attributes_to_speak()
        presented = False

        for ax_text_attr in attr_list:
            key = ax_text_attr.get_attribute_name()
            value = ax_text_attr.get_value_from_attrs(attrs)
            if not ax_text_attr.value_is_default(value):
                presentation_manager.get_manager().present_message(
                    self._localize_text_attribute(key, value),
                )
                presented = True

        if not presented:
            msg = (
                messages.CHARACTER_FORMATTING_DEFAULT
                if attrs
                else messages.CHARACTER_FORMATTING_NOT_AVAILABLE
            )
            presentation_manager.get_manager().present_message(
                msg,
            )

        return True

    @dbus_service.command
    def show_character_attributes(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Shows the font and formatting details for the current character."""

        tokens = [
            "WHERE AM I PRESENTER: show_character_attributes. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        attrs, char = self._get_current_character_info(script)
        descriptions = self._get_all_available_text_attributes(attrs)
        if not descriptions:
            presentation_manager.get_manager().present_message(
                messages.CHARACTER_FORMATTING_NOT_AVAILABLE,
            )
            return True

        title = guilabels.CHARACTER_FORMATTING
        if char:
            title = guilabels.CHARACTER_FORMATTING_FOR % char
        text = self._get_character_formatting_text(attrs)
        self._char_attributes_gui = CharacterAttributesGUI(title, text)
        self._char_attributes_gui.show_gui()

        return True

    @dbus_service.command
    def present_size_and_position(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the size and position of the current object."""

        tokens = [
            "WHERE AM I PRESENTER: present_size_and_position. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        if flat_review_presenter.get_presenter().is_active():
            obj = flat_review_presenter.get_presenter().get_current_object(script, event)
        else:
            obj = focus_manager.get_manager().get_locus_of_focus()

        rect = AXComponent.get_rect(obj)
        if AXUtilities.is_empty_rect(rect):
            full = messages.LOCATION_NOT_FOUND_FULL
            brief = messages.LOCATION_NOT_FOUND_BRIEF
            presentation_manager.get_manager().present_message(full, brief)
            return True

        full = messages.SIZE_AND_POSITION_FULL % (rect.width, rect.height, rect.x, rect.y)
        brief = messages.SIZE_AND_POSITION_BRIEF % (rect.width, rect.height, rect.x, rect.y)
        presentation_manager.get_manager().present_message(full, brief)
        return True

    @dbus_service.command
    def present_title(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the title of the current window."""

        tokens = [
            "WHERE AM I PRESENTER: present_title. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        obj = focus_manager.get_manager().get_locus_of_focus()
        if AXObject.is_dead(obj):
            obj = focus_manager.get_manager().get_active_window()

        if obj is None or AXObject.is_dead(obj):
            presentation_manager.get_manager().present_message(messages.LOCATION_NOT_FOUND_FULL)
            return True

        presentation_manager.get_manager().present_window_title(script, obj)
        return True

    @dbus_service.command
    def present_default_button(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        dialog: Atspi.Accessible | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the default button of the current dialog."""

        tokens = [
            "WHERE AM I PRESENTER: present_default_button. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        obj = focus_manager.get_manager().get_locus_of_focus()
        if dialog is None:
            _frame, dialog = script.utilities.frame_and_dialog(obj)
        if dialog is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.DIALOG_NOT_IN_A)
            return True

        button = AXUtilities.get_default_button(dialog)
        if button is None:
            if notify_user:
                presentation_manager.get_manager().present_message(
                    messages.DEFAULT_BUTTON_NOT_FOUND,
                )
            return True

        name = AXObject.get_name(button)
        if not AXUtilities.is_sensitive(button):
            presentation_manager.get_manager().present_message(
                messages.DEFAULT_BUTTON_IS_GRAYED % name,
            )
            return True

        presentation_manager.get_manager().present_message(messages.DEFAULT_BUTTON_IS % name)
        return True

    @dbus_service.command
    def present_cell_formula(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the formula associated with the current spreadsheet cell."""

        tokens = [
            "WHERE AM I PRESENTER: present_cell_formula. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        focus = focus_manager.get_manager().get_locus_of_focus()
        cell = AXUtilities.find_ancestor_inclusive(focus, AXUtilities.is_table_cell)
        if cell is None:
            if notify_user:
                presentation_manager.get_manager().present_message(messages.TABLE_NOT_IN_A)
            return True

        text = AXUtilities.get_cell_formula(cell)
        if not text:
            text = AXText.get_all_text(cell) or AXText.get_all_text(focus) or messages.EMPTY
        if notify_user:
            presentation_manager.get_manager().present_message(text)

        return True

    @dbus_service.command
    def present_status_bar(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the status bar and info bar of the current window."""

        tokens = [
            "WHERE AM I PRESENTER: present_status_bar. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        obj = focus_manager.get_manager().get_locus_of_focus()
        frame, _dialog = script.utilities.frame_and_dialog(obj)
        if frame:
            statusbar = AXUtilities.get_status_bar(frame)
            if statusbar:
                presentation_manager.get_manager().interrupt_if_needed_for_object_presentation()
                script.present_object(statusbar)
            else:
                full = messages.STATUS_BAR_NOT_FOUND_FULL
                brief = messages.STATUS_BAR_NOT_FOUND_BRIEF
                presentation_manager.get_manager().present_message(full, brief)

            infobar = AXUtilities.get_info_bar(frame)
            if infobar and AXUtilities.is_showing_and_visible(infobar):
                script.present_object(infobar)

        return True

    @dbus_service.command
    def present_link(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents details about the current link."""

        tokens = [
            "WHERE AM I PRESENTER: present_link. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        link = focus_manager.get_manager().get_locus_of_focus()
        if not script.utilities.is_link(link):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.NOT_ON_A_LINK)
            return True

        return self._do_where_am_i(script, True, link)

    def _get_all_selected_text(self, script: default.Script, obj: Atspi.Accessible) -> str:
        """Returns the selected text of obj plus any adjacent text objects."""

        string = AXUtilities.get_selected_text(obj)[0]
        if AXUtilities.is_spreadsheet_cell(obj):
            return string

        prev_obj = script.utilities.find_previous_object(obj)
        while prev_obj:
            selection = AXUtilities.get_selected_text(prev_obj)[0]
            if not selection:
                break
            string = f"{selection} {string}"
            prev_obj = script.utilities.find_previous_object(prev_obj)

        next_obj = script.utilities.find_next_object(obj)
        while next_obj:
            selection = AXUtilities.get_selected_text(next_obj)[0]
            if not selection:
                break
            string = f"{string} {selection}"
            next_obj = script.utilities.find_next_object(next_obj)

        return string

    @dbus_service.command
    def present_selected_text(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the selected text."""

        tokens = [
            "WHERE AM I PRESENTER: present_selected_text. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        obj = focus_manager.get_manager().get_locus_of_focus()
        if obj is None:
            presentation_manager.get_manager().speak_message(messages.LOCATION_NOT_FOUND_FULL)
            return True

        text = self._get_all_selected_text(script, obj)
        if not text:
            presentation_manager.get_manager().speak_message(messages.NO_SELECTED_TEXT)
            return True

        manager = speech_presenter.get_presenter()
        indentation = manager.get_indentation_description(text, only_if_changed=False)
        text = manager.adjust_for_presentation(obj, text)
        msg = messages.SELECTED_TEXT_IS % f"{indentation} {text}"
        presentation_manager.get_manager().speak_message(msg)
        return True

    @dbus_service.command
    def present_selection(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents the selected text or selected objects."""

        tokens = [
            "WHERE AM I PRESENTER: present_selection. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        obj = focus_manager.get_manager().get_locus_of_focus()
        if obj is None:
            if not script.utilities.is_link(obj):
                presentation_manager.get_manager().speak_message(messages.LOCATION_NOT_FOUND_FULL)
            return True

        tokens = ["WHERE AM I PRESENTER: presenting selection for", obj]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        spreadsheet = AXUtilities.find_ancestor(obj, AXUtilities.is_spreadsheet_table)
        if spreadsheet is not None and script.utilities.speak_selected_cell_range(spreadsheet):
            return True

        container = AXUtilities.get_selection_container(obj)
        if container is None:
            tokens = ["WHERE AM I PRESENTER: Selection container not found for", obj]
            debug.print_tokens(debug.LEVEL_INFO, tokens, True)
            return self.present_selected_text(script, event, obj)

        selected_count = AXUtilities.selected_child_count(container)
        child_count = AXUtilities.selectable_child_count(container)
        presentation_manager.get_manager().present_message(
            messages.selected_items_count(selected_count, child_count),
        )
        if not selected_count:
            return True

        selected_items = AXUtilities.selected_children(container)
        item_names = ",".join(map(AXObject.get_name, selected_items))
        presentation_manager.get_manager().speak_message(item_names)
        return True

    def _do_where_am_i(
        self,
        script: default.Script,
        basic_only: bool = True,
        obj: Atspi.Accessible | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents details about the current location at the specified level."""

        presenter = spellcheck_presenter.get_presenter()
        if presenter.is_active():
            presenter.present_error_details(not basic_only, script)

        if math_navigator.get_navigator().is_active():
            if basic_only:
                speech = math_presenter.get_presenter().get_where_am_i()
            else:
                speech = math_presenter.get_presenter().get_where_am_i_all()
            if speech:
                presentation_manager.get_manager().present_message(speech)
            return True

        if obj is None:
            obj = focus_manager.get_manager().get_locus_of_focus()
        if AXObject.is_dead(obj):
            obj = focus_manager.get_manager().get_active_window()

        if obj is None or AXObject.is_dead(obj):
            if notify_user:
                presentation_manager.get_manager().present_message(messages.LOCATION_NOT_FOUND_FULL)
            return True

        if basic_only:
            reason = PresentationReason.WHERE_AM_I_BASIC
        else:
            reason = PresentationReason.WHERE_AM_I_DETAILED

        def real_object(acc: Atspi.Accessible) -> Atspi.Accessible:
            if AXUtilities.is_focused(acc):
                return acc

            def pred(x):
                return AXUtilities.is_table_cell_or_header(x) or AXUtilities.is_list_item(x)

            ancestor = AXUtilities.find_ancestor(acc, pred)
            if ancestor is not None and not AXUtilities.is_layout_only(
                AXObject.get_parent(ancestor),
            ):
                acc = ancestor

            return acc

        script.present_object(
            real_object(obj),
            reason=reason,
        )

        return True

    @dbus_service.command
    def where_am_i_basic(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents basic information about the current location."""

        tokens = [
            "WHERE AM I PRESENTER: where_am_i_basic. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)
        return self._do_where_am_i(script, notify_user=notify_user)

    @dbus_service.command
    def where_am_i_detailed(
        self,
        script: default.Script,
        event: input_event.InputEvent | None = None,
        notify_user: bool = True,
    ) -> bool:
        """Presents detailed information about the current location."""

        tokens = [
            "WHERE AM I PRESENTER: where_am_i_detailed. Script:",
            script,
            "Event:",
            event,
            "notify_user:",
            notify_user,
        ]
        debug.print_tokens(debug.LEVEL_INFO, tokens, True)

        # TODO - JD: For some reason, we are starting the basic where am I
        # in response to the first click. Then we do the detailed one in
        # response to the second click. Until that's fixed, interrupt the
        # first one.
        presentation_manager.get_manager().interrupt_presentation()
        return self._do_where_am_i(script, False, notify_user=notify_user)


class CharacterAttributesGUI:
    """Presents character attributes in a text view."""

    def __init__(self, title: str, text: str) -> None:
        self._gui: Gtk.Dialog = self._create_dialog(title, text)

    def _create_dialog(self, title: str, text: str) -> Gtk.Dialog:
        """Creates the dialog."""

        dialog = Gtk.Dialog(
            title,
            None,
            Gtk.DialogFlags.MODAL,
            (Gtk.STOCK_CLOSE, Gtk.ResponseType.CLOSE),
        )
        dialog.set_default_size(600, 400)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.set_hexpand(True)
        scrolled_window.set_vexpand(True)

        textbuffer = Gtk.TextBuffer()
        textbuffer.set_text(text)
        textbuffer.place_cursor(textbuffer.get_start_iter())

        textview = Gtk.TextView(buffer=textbuffer)
        textview.set_editable(False)
        textview.set_cursor_visible(True)
        textview.set_wrap_mode(Gtk.WrapMode.WORD)
        scrolled_window.add(textview)  # pylint: disable=no-member
        dialog.get_content_area().pack_start(scrolled_window, True, True, 0)
        dialog.set_focus(textview)
        dialog.connect("response", self.on_response)
        dialog.connect("key-press-event", self.on_key_press)
        return dialog

    def on_response(self, _dialog: Gtk.Dialog, response: Gtk.ResponseType) -> None:
        """Handler for the 'response' signal of the dialog."""

        if response == Gtk.ResponseType.CLOSE:
            self._gui.destroy()

    def on_key_press(self, _dialog: Gtk.Dialog, event: Gdk.EventKey) -> bool:
        """Handler for the 'key-press-event' signal of the dialog."""

        if event.keyval == Gdk.KEY_Escape:
            self._gui.destroy()
            return True
        return False

    def show_gui(self) -> None:
        """Shows the dialog."""

        self._gui.show_all()  # pylint: disable=no-member
        self._gui.present_with_time(time.time())


_presenter = WhereAmIPresenter()


def get_presenter() -> WhereAmIPresenter:
    """Returns the Where Am I Presenter"""

    return _presenter
